#!/usr/bin/env python

import sys
import os
import warnings
import numpy as np
from pandas import DataFrame, read_csv
import timeit
from xarray import open_dataset
from netCDF4 import Dataset, num2date, date2num
from mpi4py import MPI
from cfunits import Units
from numpy.ma.core import MaskError
from copy import deepcopy, copy


class Nes(object):
    def __init__(self, comm=None, path=None, info=False, dataset=None, xarray=False):
        """
        Initialize the goddess class
        """
        if comm is None:
            self.comm = MPI.COMM_WORLD
        else:
            self.comm = comm
        self.rank = self.comm.Get_rank()
        self.master = self.rank == 0
        self.size = self.comm.Get_size()

        self.print_info = info
        self.is_xarray = xarray
        self.__ini_path = path
        if dataset is not None:
            if self.is_xarray:
                self.dataset = dataset
                self.netcdf = None
            else:
                self.dataset = None
                self.netcdf = dataset
        elif self.__ini_path is not None:
            if self.is_xarray:
                self.dataset = self.open_dataset()
                self.netcdf = None
            else:
                self.dataset = None
                self.netcdf = self.open_netcdf4()

        self.variables = self.get_lazy_variables()

        self.time = self.__get_time()
        # self.time = self.get_coordinate_values(self.__time)
        self._lat = self.get_coordinate_dimension(['lat', 'latitude'])
        self._lon = self.get_coordinate_dimension(['lon', 'longitude'])
        self._lev = self.get_coordinate_dimension(['lev', 'level', 'lm'])

        y_len = self._lat['data'].shape[0]
        self.y_min = (y_len // self.size) * self.rank
        if self.rank + 1 == self.size:
            self.y_max = None
        else:
            self.y_max = (y_len // self.size) * (self.rank + 1)

        self.lat = self.get_coordinate_values(self._lat)
        self.lon = self.get_coordinate_values(self._lon)
        self.lev = self.get_coordinate_values(self._lev)

        self.global_attrs = self.get_global_attributes()
        # else:
        #     self.dataset = None
        #     self.netcdf = None
        #     self.variables = {}
        #     self.time = None
        #     self.__lat = None
        #     self.lat = None
        #     self.__lon = None
        #     self.lon = None
        #     self.__lev = None
        #     self.lev = None
        #     self.y_min = None
        #     self.y_max = None
        self.t_min = 0

        self.zip_lvl = 0

        self.var_dim = ('lat', 'lon')
        self.lat_dim = ('lat',)
        self.lon_dim = ('lon',)

    def __getstate__(self):
        d = self.__dict__
        self_dict = {k: d[k] for k in d if k not in ['comm', 'variables', 'netcdf']}
        return self_dict

    def __setstate__(self, state):
        self.__dict__ = state
        return None

    def copy(self, copy_vars=False):
        # self.close()
        new = deepcopy(self)
        new.netcdf = None
        # new.open()
        if copy_vars:
            new.variables = new.get_lazy_variables()
        else:
            new.variables = {}
        return new

    def clear_communicator(self):
        self.comm = None
        self.rank = 0
        self.master = 0
        self.size = 0

        self.y_min = 0
        self.y_max = 0

        return None

    def set_communicator(self, comm):
        self.comm = comm
        self.rank = self.comm.Get_rank()
        self.master = self.rank == 0
        self.size = self.comm.Get_size()

        # Update distribution
        y_len = self._lat['data'].shape[0]
        self.y_min = (y_len // self.size) * self.rank
        if self.rank + 1 == self.size:
            self.y_max = None
        else:
            self.y_max = (y_len // self.size) * (self.rank + 1)
        return None

    def open(self):
        if self.is_xarray:
            self.dataset = self.open_dataset()
            self.netcdf = None
        else:
            self.dataset = None
            self.netcdf = self.open_netcdf4()
        return None

    def open_dataset(self):
        if self.master:
            warnings.filterwarnings('ignore')  # Disabling warnings while reading MONARCH original file
            dataset = open_dataset(self.__ini_path, decode_coords='all')
            warnings.filterwarnings('default')  # Re-activating warnings
        else:
            dataset = None
        dataset = self.comm.bcast(dataset, root=0)
        self.dataset = dataset
        return dataset

    def open_netcdf4(self, mode='r'):
        """
        Open a NetCDF file to be read in the future.

        :param mode: Indicates the mode to open the netCDF. DEFAULT = 'r' (read)
        :type mode: str

        :return: NetCDF object
        :rtype: Dataset
        """
        if self.size == 1:
            netcdf = Dataset(self.__ini_path, format="NETCDF4", mode=mode, parallel=False)
        else:
            netcdf = Dataset(self.__ini_path, format="NETCDF4", mode=mode, parallel=True, comm=self.comm,
                             info=MPI.Info())
        self.netcdf = netcdf
        return netcdf

    def close(self):
        if self.netcdf is not None:
            self.netcdf.close()
            self.netcdf = None
        return None

    @staticmethod
    def __parse_time_unit(t_units):
        """
        Parses the time units to be CF compliant

        :param t_units: Original time units.
        :type t_units: str

        :return: CF compliant time units.
        :rtype: str
        """
        if 'h @' in t_units:
            t_units = 'hour since {0}-{1}-{2} {3}:{4}:{5} UTC'.format(
                t_units[4:8], t_units[8:10], t_units[10:12], t_units[13:15], t_units[15:17], t_units[17:-4])
        return t_units

    def __get_time(self):
        if self.is_xarray:
            time = self.variables['time']
        else:
            if self.master:
                nc_var = self.netcdf.variables['time']
                time = num2date(nc_var[:], self.__parse_time_unit(nc_var.units), calendar=nc_var.calendar)
                time = [aux.replace(second=0, microsecond=0) for aux in time]
            else:
                time = None
            time = self.comm.bcast(time, root=0)
        self.free_vars('time')
        return time

    def get_coordinate_dimension(self, possible_names):
        if isinstance(possible_names, str):
            possible_names = [possible_names]
        dimension_name = set(possible_names).intersection(set(self.variables.keys())).pop()

        if self.is_xarray:
            nc_var = self.dataset[dimension_name]
        else:
            nc_var = self.variables[dimension_name].copy()
            nc_var['data'] = self.netcdf.variables[dimension_name][:]

        self.free_vars(dimension_name)
        return nc_var

    def get_coordinate_values(self, coordinate_info):
        coordinate_len = len(coordinate_info['data'].shape)
        values = deepcopy(coordinate_info)
        if coordinate_len == 2:
            values['data'] = values['data'][self.y_min:self.y_max]
        return values

    def get_lazy_variables(self):
        if self.is_xarray:
            variables = self.dataset.variables
        else:
            if self.master:
                variables = {}
                for var_name, var_info in self.netcdf.variables.items():
                    variables[var_name] = {}
                    variables[var_name]['data'] = None
                    for attrname in var_info.ncattrs():
                        # Avoiding some attributes
                        if attrname not in ['missing_value', '_FillValue']:
                            value = getattr(var_info, attrname)
                            if value == 'unitless':
                                value = ''
                            variables[var_name][attrname] = value
            else:
                variables = None
            variables = self.comm.bcast(variables, root=0)

        return variables

    def read_variable(self, var_name):
        nc_var = self.netcdf.variables[var_name]
        var_dims = nc_var.dimensions
        if len(var_dims) == 2:
            data = nc_var[self.y_min:self.y_max, :]
        elif len(var_dims) == 3:
            data = nc_var[self.t_min::, self.y_min:self.y_max, :]
        elif len(var_dims) == 4:
            data = nc_var[self.t_min::, :, self.y_min:self.y_max, :]
        elif len(var_dims) == 5:
            data = nc_var[self.t_min::, :, :, self.y_min:self.y_max, :]
        else:
            raise NotImplementedError('Only can be read netCDF with 5 dimensions')

        # Missing to nan
        try:
            data[data.mask == True] = np.nan
        except (AttributeError, MaskError):
            pass
        return data

    def load(self, varlist=None):
        if self.netcdf is None:
            self.open_dataset()
            close = True
        else:
            close = False

        if varlist is None:
            varlist = list(self.variables.keys())

        for i, var_name in enumerate(varlist):
            if self.print_info:
                print("Rank {0:03d}: Loading {1} var ({2}/{3})".format(self.rank, var_name, i + 1, len(varlist)))
            self.variables[var_name]['data'] = self.read_variable(var_name)
            if self.print_info:
                print("Rank {0:03d}: Loaded {1} var ({2})".format(self.rank, var_name, self.variables[var_name]['data'].shape))

        if close:
            self.close()
        return True

    def get_global_attributes(self):
        gl_attrs = {}
        if self.is_xarray:
            gl_attrs = self.dataset.attrs
        else:
            for attrname in self.netcdf.ncattrs():
                gl_attrs[attrname] = getattr(self.netcdf, attrname)
        return gl_attrs

    def free_vars(self, var_list):
        if isinstance(var_list, str):
            var_list = [var_list]

        if self.is_xarray:
            self.dataset = self.dataset.drop_vars(var_list)
            self.variables = self.get_lazy_variables()
        else:
            for var_name in var_list:
                del self.variables[var_name]
        return True

    def keep_vars(self, var_list):
        if isinstance(var_list, str):
            var_list = [var_list]

        to_remove = list(set(self.variables.keys()).difference(set(var_list)))

        self.free_vars(to_remove)
        return True

    def create_dimensions(self, netcdf):
        """
        Create the 'lev' and 'time' dimension.

        :param netcdf: NetCDF object.
        :type netcdf: Dataset

        :return: True when the writing is finished.
        :rtype: bool
        """
        # netcdf.createDimension('time', None)
        netcdf.createDimension('time', len(self.time))
        netcdf.createDimension('lev', len(self._lev['data']))

        return True

    def create_dimension_variables(self, netcdf):
        """
        Create the 'lev', 'time', 'lat' and 'lon' variables.

        :param netcdf: NetCDF object.
        :type netcdf: Dataset

        :return: True when the writing is finished.
        :rtype: bool
        """
        # TIMES
        time_var = netcdf.createVariable('time', np.float64, ('time',), zlib=self.zip_lvl > 0,
                                         complevel=self.zip_lvl)
        time_var.units = 'hours since {0}'.format(self.time[0].strftime("%Y-%m-%d %H:%M:%S"))
        time_var.standard_name = "time"
        time_var.calendar = 'standard'
        time_var.long_name = "time"
        if self.size > 1:
            time_var.set_collective(True)
        time_var[:] = date2num(self.time, time_var.units, calendar='standard')

        # LEVELS
        lev = netcdf.createVariable('lev', np.float64, ('lev',), zlib=self.zip_lvl > 0, complevel=self.zip_lvl)
        lev.units = Units(self._lev['units'], formatted=True).units
        lev.positive = 'up'
        if self.size > 1:
            lev.set_collective(True)
        lev[:] = self._lev['data']

        # LATITUDES
        lats = netcdf.createVariable('lat', np.float64, self.lat_dim, zlib=self.zip_lvl > 0, complevel=self.zip_lvl)
        lats.units = "degrees_north"
        lats.axis = "Y"
        lats.long_name = "latitude coordinate"
        lats.standard_name = "latitude"
        if self.size > 1:
            lats.set_collective(True)
        lats[:] = self._lat['data']

        # LONGITUDES
        lons = netcdf.createVariable('lon', np.float64, self.lon_dim, zlib=self.zip_lvl > 0, complevel=self.zip_lvl)
        lons.units = "degrees_east"
        lons.axis = "X"
        lons.long_name = "longitude coordinate"
        lons.standard_name = "longitude"
        if self.size > 1:
            lons.set_collective(True)
        lons[:] = self._lon['data']

        return None

    def create_variables(self, netcdf, chunking=False):
        for i, (var_name, var_dict) in enumerate(self.variables.items()):
            if self.print_info:
                print("Rank {0:03d}: Writing {1} var ({2}/{3})".format(self.rank, var_name, i + 1, len(self.variables)))
            try:
                if not chunking:
                    var = netcdf.createVariable(var_name, var_dict['data'].dtype, ('time', 'lev',) + self.var_dim,
                                                zlib=self.zip_lvl > 0, complevel=self.zip_lvl)
                else:
                    if self.master:
                        chunk_size = var_dict['data'].shape
                    else:
                        chunk_size = None
                    chunk_size = self.comm.bcast(chunk_size, root=0)
                    var = netcdf.createVariable(var_name, var_dict['data'].dtype, ('time', 'lev',) + self.var_dim,
                                                zlib=self.zip_lvl > 0, complevel=self.zip_lvl, chunksizes=chunk_size)
                if self.print_info:
                    print("Rank {0:03d}: Var {1} created ({2}/{3})".format(self.rank, var_name, i + 1, len(self.variables)))
                if self.size > 1:
                    var.set_collective(True)
                    if self.print_info:
                        print("Rank {0:03d}: Var {1} collective ({2}/{3})".format(self.rank, var_name, i + 1, len(self.variables)))

                for att_name, att_value in var_dict.items():
                    if att_name == 'data':
                        try:
                            var[:, :, self.y_min:self.y_max, :] = att_value
                        except ValueError:
                            var[:, 0, self.y_min:self.y_max, :] = att_value
                            # msg = "*WARNING* '{0}' variable is a 3D field. Setting it on first (0) layer.".format(
                            #     var_name)
                            # warn(msg)
                        if self.print_info:
                            print("Rank {0:03d}: Var {1} data ({2}/{3})".format(self.rank, var_name, i + 1,
                                                                                len(self.variables)))
                    elif att_name not in ['chunk_size', 'var_dims']:
                        var.setncattr(att_name, att_value)
                if self.print_info:
                    print("Rank {0:03d}: Var {1} completed ({2}/{3})".format(self.rank, var_name, i + 1,
                                                                             len(self.variables)))
                # self.set_var_crs(var)
            except Exception as e:
                print("**ERROR** an error hase occur while writing the '{0}' variable".format(var_name))
                # print("**ERROR** an error hase occur while writing the '{0}' variable".format(var_name),
                #       file=sys.stderr)
                raise e

        return True

    def set_crs(self, netcdf):
        """
        Must be implemented on inner class

        :param netcdf: NetCDF object.
        :type netcdf: Dataset

        :return: True when the writing is finished.
        :rtype: bool
        """
        return None

    def to_netcdf_py(self, path, chunking=False):
        # Open NetCDF
        if self.print_info:
            print("Rank {0:03d}: Creating {1}".format(self.rank, path))
        if self.size > 1:
            netcdf = Dataset(path, format="NETCDF4", mode='w', parallel=True, comm=self.comm, info=MPI.Info())
        else:
            netcdf = Dataset(path, format="NETCDF4", mode='w', parallel=False)
        if self.print_info:
            print("Rank {0:03d}: NetCDF ready to write".format(self.rank))
        # Create Dimensions
        self.create_dimensions(netcdf)
        # Create dimension variables
        self.create_dimension_variables(netcdf)
        if self.print_info:
            print("Rank {0:03d}: Dimensions done".format(self.rank))

        # Create variables
        self.create_variables(netcdf, chunking=chunking)

        # Create metadata
        self.set_crs(netcdf)

        # Close NetCDF
        if self.global_attrs is not None:
            for att_name, att_value in self.global_attrs.items():
                netcdf.setncattr(att_name, att_value)
        netcdf.setncattr('Conventions', 'CF-1.7')

        self.comm.Barrier()
        netcdf.close()

        return None

    def to_netcdf(self, path, compression_level=0, serial=False, info=False, chunking=False):
        old_info = self.print_info
        self.print_info = info

        self.zip_lvl = compression_level
        if self.is_xarray:
            pass
        else:
            # if serial:
            if serial and self.size > 1:
                data = self.gather_data()
                if self.master:
                    new_nc = self.copy(copy_vars=False)
                    new_nc.set_communicator(MPI.COMM_SELF)
                    new_nc.variables = data
                    new_nc.to_netcdf_py(path)
            else:
                self.to_netcdf_py(path, chunking=chunking)

        self.print_info = old_info

        return None

    def gather_data(self):
        data_list = deepcopy(self.variables)
        for var_name in data_list.keys():
            try:
                # noinspection PyArgumentList
                data_aux = self.comm.gather(data_list[var_name]['data'], root=0)
                if self.rank == 0:
                    if len(data_list[var_name]['data'].shape) == 2:
                        axis = 0
                    elif len(data_list[var_name]['data'].shape) == 3:
                        axis = 1
                    else:
                        axis = 2
                    data_list[var_name]['data'] = np.concatenate(data_aux, axis=axis)
            except Exception as e:
                print("**ERROR** an error hase occur while gathering the '{0}' variable".format(var_name))
                sys.stderr.write("**ERROR** an error hase occur while gathering the '{0}' variable".format(var_name))
                print(e)
                sys.stderr.write(str(e))
                # print(e, file=sys.stderr)
                sys.stderr.flush()
                self.comm.Abort(1)
                raise e

        return data_list
