#!/usr/bin/env python

from nes.nc_projections.default_nes import Nes
from cfunits import Units


class RotatedNes(Nes):
    def __init__(self, comm=None, path=None, info=False, dataset=None, xarray=False):
        """
        Initialize the goddess class
        """
        super(RotatedNes, self).__init__(comm=comm, path=path, info=info, dataset=dataset, xarray=xarray)

        self._rlat = self.get_coordinate_dimension('rlat')
        self._rlon = self.get_coordinate_dimension('rlon')

        self.rlat = self.get_coordinate_values(self._rlat)
        self.rlon = self.get_coordinate_values(self._rlon)

        self.projection_data = self.get_projection_data()

        self.var_dim = ('rlat', 'rlon')
        self.lat_dim = ('rlat', 'rlon')
        self.lon_dim = ('rlat', 'rlon')

    def get_projection_data(self):
        projection = self.variables['rotated_pole']
        self.free_vars('rotated_pole')
        return projection

    def create_dimensions(self, netcdf):
        """
        Create the 'rlat', 'rlon' dimensions and the super dimensions ('lev', 'time').

        :param netcdf: NetCDF object.
        :type netcdf: Dataset

        :return: True when the writing is finished.
        :rtype: bool
        """
        super(RotatedNes, self).create_dimensions(netcdf)

        netcdf.createDimension('rlon', len(self._rlon['data']))
        netcdf.createDimension('rlat', len(self._rlat['data']))

        return None

    def create_dimension_variables(self, netcdf):
        """
        Create the 'rlat' and 'rlon' variables.

        :param netcdf: NetCDF object.
        :type netcdf: Dataset

        :return: True when the writing is finished.
        :rtype: bool
        """

        super(RotatedNes, self).create_dimension_variables(netcdf)

        rlat = netcdf.createVariable('rlat', self._rlat['data'].dtype, ('rlat',))
        rlat.long_name = "latitude in rotated pole grid"
        rlat.units = Units("degrees", formatted=True).units
        rlat.standard_name = "grid_latitude"
        if self.size > 1:
            rlat.set_collective(True)
        rlat[:] = self._rlat['data']

        # Rotated Longitude
        rlon = netcdf.createVariable('rlon', self._rlon['data'].dtype, ('rlon',))
        rlon.long_name = "longitude in rotated pole grid"
        rlon.units = Units("degrees", formatted=True).units
        rlon.standard_name = "grid_longitude"
        if self.size > 1:
            rlon.set_collective(True)
        rlon[:] = self._rlon['data']

        return None

    @staticmethod
    def set_var_crs(var):
        """
        Set the grid_mapping to 'rotated_pole'.

        :param var: netCDF var object.
        :type var: Variable

        :return: True when the writing is finished.
        :rtype: bool
        """
        var.grid_mapping = 'rotated_pole'

        return None

    def set_crs(self, netcdf):
        """
        Create the 'crs' variable for the rotated latitude longitude grid_mapping.

        :param netcdf: NetCDF object.
        :type netcdf: Dataset

        :return: True when the writing is finished.
        :rtype: bool
        """

        mapping = netcdf.createVariable('rotated_pole', 'c')
        mapping.grid_mapping_name = 'rotated_latitude_longitude'
        mapping.grid_north_pole_latitude = self.projection_data['grid_north_pole_latitude']
        mapping.grid_north_pole_longitude = self.projection_data['grid_north_pole_longitude']

        return None
